## ----eval=FALSE---------------------------------------------------------------
# if (!require("BiocManager", quietly=TRUE))
#     install.packages("BiocManager")
# 
# pkgs <- c("HDF5Array", "ExperimentHub", "DelayedMatrixStats", "RSpectra")
# BiocManager::install(pkgs)

## ----message=FALSE------------------------------------------------------------
library(HDF5Array)
library(ExperimentHub)
library(DelayedMatrixStats)
library(RSpectra)

## -----------------------------------------------------------------------------
hub <- ExperimentHub()
hub["EH1039"]$description  # sparse representation
hub["EH1040"]$description  # dense representation

## ----message=FALSE------------------------------------------------------------
## Note that this will be quick if the HDF5 files are already in the
## local ExperimentHub cache. Otherwise, it will take a while
full_sparse_h5 <- hub[["EH1039"]]
full_dense_h5  <- hub[["EH1040"]]

## -----------------------------------------------------------------------------
## Use 'h5ls(full_sparse_h5)' to find out the group.
full_sparse <- TENxMatrix(full_sparse_h5, group="mm10")
class(full_sparse)
dim(full_sparse)

## -----------------------------------------------------------------------------
## Use 'h5ls(full_dense_h5)' to find out the name of the dataset.
full_dense <- HDF5Array(full_dense_h5, name="counts")
class(full_dense)
dim(full_dense)

## -----------------------------------------------------------------------------
dimnames(full_dense) <- dimnames(full_sparse)

## -----------------------------------------------------------------------------
sparse1 <- full_sparse[ , 1:12500]
dense1  <- full_dense[ , 1:12500]

sparse2 <- full_sparse[ , 1:25000]
dense2  <- full_dense[ , 1:25000]

sparse3 <- full_sparse[ , 1:50000]
dense3  <- full_dense[ , 1:50000]

sparse4 <- full_sparse[ , 1:100000]
dense4  <- full_dense[ , 1:100000]

sparse5 <- full_sparse[ , 1:200000]
dense5  <- full_dense[ , 1:200000]

## -----------------------------------------------------------------------------
## Keep 1000 most variable genes by default.
simple_normalize <- function(mat, num_variable_genes=1000)
{
    stopifnot(length(dim(mat)) == 2, !is.null(rownames(mat)))
    mat <- mat[rowSums(mat) > 0, ]
    mat <- t(t(mat) * 10000 / colSums(mat))
    row_vars <- rowVars(mat)
    rv_order <- order(row_vars, decreasing=TRUE)
    variable_idx <- head(rv_order, n=num_variable_genes)
    mat <- log1p(mat[variable_idx, ])
    mat / rowSds(mat)
}

## -----------------------------------------------------------------------------
simple_PCA <- function(mat, k=25)
{
    stopifnot(length(dim(mat)) == 2)
    row_means <- rowMeans(mat)
    Ax <- function(x, args)
        (as.numeric(mat %*% x) - row_means * sum(x))
    Atx <- function(x, args)
        (as.numeric(x %*% mat) - as.vector(row_means %*% x))
    RSpectra::svds(Ax, Atrans=Atx, k=k, dim=dim(mat))
}

## -----------------------------------------------------------------------------
DelayedArray::setAutoBlockSize(2.5e8)  # blocks of 250 Mb

## ----normalize_sparse1--------------------------------------------------------
dim(sparse1)
system.time(sparse1n <- simple_normalize(sparse1))
gc()

## ----results="hide"-----------------------------------------------------------
dim(sparse1n)
sparse1n_path <- tempfile()
writeTENxMatrix(sparse1n, sparse1n_path, group="matrix", level=0)

## ----normalize_sparse2--------------------------------------------------------
dim(sparse2)
system.time(sparse2n <- simple_normalize(sparse2))
gc()

## ----results="hide"-----------------------------------------------------------
dim(sparse2n)
sparse2n_path <- tempfile()
writeTENxMatrix(sparse2n, sparse2n_path, group="matrix", level=0)

## -----------------------------------------------------------------------------
DelayedArray::setAutoBlockSize(1e8)  # blocks of 100 Mb

## ----normalize_dense1---------------------------------------------------------
dim(dense1)
system.time(dense1n <- simple_normalize(dense1))
gc()

## ----results="hide"-----------------------------------------------------------
dim(dense1n)
dense1n_path <- tempfile()
writeHDF5Array(dense1n, dense1n_path, name="normalized_counts", level=0)

## ----normalize_dense2---------------------------------------------------------
dim(dense2)
system.time(dense2n <- simple_normalize(dense2))
gc()

## ----results="hide"-----------------------------------------------------------
dim(dense2n)
dense2n_path <- tempfile()
writeHDF5Array(dense2n, dense2n_path, name="normalized_counts", level=0)

## -----------------------------------------------------------------------------
DelayedArray::setAutoBlockSize(1e8)  # blocks of 100 Mb

## ----PCA_sparse1n-------------------------------------------------------------
sparse1n <- TENxMatrix(sparse1n_path)
dim(sparse1n)
system.time(pca1s <- simple_PCA(sparse1n))
gc()

## ----PCA_sparse2n-------------------------------------------------------------
sparse2n <- TENxMatrix(sparse2n_path)
dim(sparse2n)
system.time(pca2s <- simple_PCA(sparse2n))
gc()

## -----------------------------------------------------------------------------
DelayedArray::setAutoBlockSize(1e8)  # blocks of 100 Mb

## ----PCA_dense1n--------------------------------------------------------------
dense1n <- HDF5Array(dense1n_path, name="normalized_counts")
dim(dense1n)
system.time(pca1d <- simple_PCA(dense1n))
gc()

## ----PCA_dense2n--------------------------------------------------------------
dense2n <- HDF5Array(dense2n_path, name="normalized_counts")
dim(dense2n)
system.time(pca2d <- simple_PCA(dense2n))
gc()

## ----sanity_checks------------------------------------------------------------
stopifnot(all.equal(pca1s, pca1d))
stopifnot(all.equal(pca2s, pca2d))

## -----------------------------------------------------------------------------
sessionInfo()

